using System;
using System.Text;
using System.Runtime.InteropServices;
using System.Runtime.InteropServices.ComTypes;

namespace BoxedAppSDK
{
    internal sealed class NativeMethods
    {
		[DllImport("BoxedAppSDK.dll")]
        internal static extern bool BoxedAppSDK_Init();

		[DllImport("BoxedAppSDK.dll")]
		internal static extern void BoxedAppSDK_Exit();

		[Flags]
			public enum EFileAccess : uint
		{
			/// <summary>
			/// 
			/// </summary>
			GenericRead = 0x80000000,
			/// <summary>
			/// 
			/// </summary>
			GenericWrite = 0x40000000,
			/// <summary>
			/// 
			/// </summary>
			GenericExecute = 0x20000000,
			/// <summary>
			/// 
			/// </summary>
			GenericAll = 0x10000000
		}

		[Flags]
			public enum EFileShare : uint
		{
			/// <summary>
			/// 
			/// </summary>
			None = 0x00000000,
			/// <summary>
			/// Enables subsequent open operations on an object to request read access. 
			/// Otherwise, other processes cannot open the object if they request read access. 
			/// If this flag is not specified, but the object has been opened for read access, the function fails.
			/// </summary>
			Read = 0x00000001,
			/// <summary>
			/// Enables subsequent open operations on an object to request write access. 
			/// Otherwise, other processes cannot open the object if they request write access. 
			/// If this flag is not specified, but the object has been opened for write access, the function fails.
			/// </summary>
			Write = 0x00000002,
			/// <summary>
			/// Enables subsequent open operations on an object to request delete access. 
			/// Otherwise, other processes cannot open the object if they request delete access.
			/// If this flag is not specified, but the object has been opened for delete access, the function fails.
			/// </summary>
			Delete = 0x00000004
		}

		public enum ECreationDisposition : uint
		{
			/// <summary>
			/// Creates a new file. The function fails if a specified file exists.
			/// </summary>
			New = 1,
			/// <summary>
			/// Creates a new file, always. 
			/// If a file exists, the function overwrites the file, clears the existing attributes, combines the specified file attributes, 
			/// and flags with FILE_ATTRIBUTE_ARCHIVE, but does not set the security descriptor that the SECURITY_ATTRIBUTES structure specifies.
			/// </summary>
			CreateAlways = 2,
			/// <summary>
			/// Opens a file. The function fails if the file does not exist. 
			/// </summary>
			OpenExisting = 3,
			/// <summary>
			/// Opens a file, always. 
			/// If a file does not exist, the function creates a file as if dwCreationDisposition is CREATE_NEW.
			/// </summary>
			OpenAlways = 4,
			/// <summary>
			/// Opens a file and truncates it so that its size is 0 (zero) bytes. The function fails if the file does not exist.
			/// The calling process must open the file with the GENERIC_WRITE access right. 
			/// </summary>
			TruncateExisting = 5
		}

		[Flags]
			public enum EFileAttributes : uint
		{
			Readonly         = 0x00000001,
			Hidden           = 0x00000002,
			System           = 0x00000004,
			Directory        = 0x00000010,
			Archive          = 0x00000020,
			Device           = 0x00000040,
			Normal           = 0x00000080,
			Temporary        = 0x00000100,
			SparseFile       = 0x00000200,
			ReparsePoint     = 0x00000400,
			Compressed       = 0x00000800,
			Offline          = 0x00001000,
			NotContentIndexed= 0x00002000,
			Encrypted        = 0x00004000,
			Write_Through    = 0x80000000,
			Overlapped       = 0x40000000,
			NoBuffering      = 0x20000000,
			RandomAccess     = 0x10000000,
			SequentialScan   = 0x08000000,
			DeleteOnClose    = 0x04000000,
			BackupSemantics  = 0x02000000,
			PosixSemantics   = 0x01000000,
			OpenReparsePoint = 0x00200000,
			OpenNoRecall     = 0x00100000,
			FirstPipeInstance= 0x00080000
		}	

		[DllImport("BoxedAppSDK.dll", SetLastError = true, CharSet = CharSet.Auto)]
		internal static extern IntPtr BoxedAppSDK_CreateVirtualFile(
			string lpFileName,
			EFileAccess dwDesiredAccess,
			EFileShare dwShareMode,
			IntPtr lpSecurityAttributes,
			ECreationDisposition dwCreationDisposition,
			EFileAttributes dwFlagsAndAttributes,
			IntPtr hTemplateFile);

		[DllImport("BoxedAppSDK.dll", SetLastError = true, CharSet = CharSet.Auto)]
		internal static extern int BoxedAppSDK_RegisterCOMLibraryInVirtualRegistry(string lpFileName);
	
        [DllImport("BoxedAppSDK.dll", SetLastError = true, CharSet = CharSet.Ansi)]
        internal static extern void BoxedAppSDK_SetContext(string code);

        [DllImport("BoxedAppSDK.dll", SetLastError = true)]
        internal static extern void BoxedAppSDK_EnableOption(UInt32 dwOptionIndex, bool bEnable);

        [DllImport("BoxedAppSDK.dll", SetLastError = true)]
        internal static extern bool BoxedAppSDK_IsOptionEnabled(UInt32 dwOptionIndex);

		[DllImport("BoxedAppSDK.dll", SetLastError = true, CharSet = CharSet.Auto)]
		internal static extern IntPtr BoxedAppSDK_CreateVirtualFileBasedOnIStream(
			string lpFileName,
			EFileAccess dwDesiredAccess,
			EFileShare dwShareMode,
			IntPtr lpSecurityAttributes,
			ECreationDisposition dwCreationDisposition,
			EFileAttributes dwFlagsAndAttributes,
			IntPtr hTemplateFile, 
			IStream Stream);

        [Flags]
        public enum RegOption
        {
            NonVolatile = 0x0,
            Volatile = 0x1,
            CreateLink = 0x2,
            BackupRestore = 0x4,
            OpenLink = 0x8
        }

        [Flags]
        public enum RegSAM
        {
            QueryValue = 0x0001,
            SetValue = 0x0002,
            CreateSubKey = 0x0004,
            EnumerateSubKeys = 0x0008,
            Notify = 0x0010,
            CreateLink = 0x0020,
            WOW64_32Key = 0x0200,
            WOW64_64Key = 0x0100,
            WOW64_Res = 0x0300,
            Read = 0x00020019,
            Write = 0x00020006,
            Execute = 0x00020019,
            AllAccess = 0x000f003f
        }

        public enum RegResult
        {
            CreatedNewKey = 0x00000001,
            OpenedExistingKey = 0x00000002
        }

        [StructLayout(LayoutKind.Sequential)]
        public struct SECURITY_ATTRIBUTES
        {
            public int nLength;
            public IntPtr lpSecurityDescriptor;
            public int bInheritHandle;
        }

        [DllImport("BoxedAppSDK.dll", SetLastError = true, CharSet = CharSet.Auto)]
        internal static extern int BoxedAppSDK_CreateVirtualRegKey(
                    int hKey,
                    string lpSubKey,
                    int Reserved,
                    string lpClass,
                    RegOption dwOptions,
                    RegSAM samDesired,
                    SECURITY_ATTRIBUTES lpSecurityAttributes,
                    out int phkResult,
                    out RegResult lpdwDisposition);

        public const UInt32 DEF_BOXEDAPPSDK_OPTION__ALL_CHANGES_ARE_VIRTUAL = 1; // default: 0 (FALSE)
        public const UInt32 DEF_BOXEDAPPSDK_OPTION__EMBED_BOXEDAPP_IN_CHILD_PROCESSES = 2; // default: 1 (FALSE, don't enable BoxedApp to a new process by default)
        public const UInt32 DEF_BOXEDAPPSDK_OPTION__ENABLE_VIRTUAL_FILE_SYSTEM = 3; // default: 1 (TRUE)
        public const UInt32 DEF_BOXEDAPPSDK_OPTION__RECREATE_VIRTUAL_FILE_AS_VIRTUAL = 4; // default: 1 (TRUE)
        public const UInt32 DEF_BOXEDAPPSDK_OPTION__ENABLE_VIRTUAL_REGISTRY = 5; // default: 1 (TRUE)
        public const UInt32 DEF_BOXEDAPPSDK_OPTION__HIDE_VIRTUAL_FILES_FROM_FILE_DIALOG = 6; // default: 0 (FALSE)

        public class AllChangesAreVirtual : IDisposable
        {
            public AllChangesAreVirtual()
            { 
                BoxedAppSDK_EnableOption(DEF_BOXEDAPPSDK_OPTION__ALL_CHANGES_ARE_VIRTUAL, true);
            }

            public void Dispose()
            {
                BoxedAppSDK_EnableOption(DEF_BOXEDAPPSDK_OPTION__ALL_CHANGES_ARE_VIRTUAL, false);
            }
        }
    }
}
