#include "StdAfx.h"
#include "MyVirtualFile.h"

class CFileProperty
{
private:
	SBoxedAppVirtualFileData m_data;

	// the name is here, not in m_data
	std::wstring m_strName;

public:
	CFileProperty(
		LPCWSTR szName, 
		LARGE_INTEGER nSize, 
		DWORD dwAttributes, 
		LARGE_INTEGER CreationTime, 
		LARGE_INTEGER LastWriteTime, 
		LARGE_INTEGER LastAccessTime, 
		LARGE_INTEGER ChangeTime)
	{
		m_data.m_dwSizeOfStruct = sizeof(m_data);
		m_strName = szName;
		m_data.m_nSize = nSize;
		m_data.m_dwAttributes = dwAttributes;

		memcpy(&m_data.m_CreationTime, &CreationTime, sizeof(LARGE_INTEGER));
		memcpy(&m_data.m_LastWriteTime, &LastWriteTime, sizeof(LARGE_INTEGER));
		memcpy(&m_data.m_LastAccessTime, &LastAccessTime, sizeof(LARGE_INTEGER));
		memcpy(&m_data.m_ChangeTime, &ChangeTime, sizeof(LARGE_INTEGER));
	}

	~CFileProperty()
	{
	}

	void CopyTo(SBoxedAppVirtualFileData& other) const
	{
		memcpy(&other, &m_data, sizeof(m_data));

		// You should use CoTaskMemAlloc to alloc memory here
		other.m_szName = (LPOLESTR)CoTaskMemAlloc(sizeof(WCHAR)* (m_strName.length() + 1));
		lstrcpyW(other.m_szName, m_strName.c_str());
	}
};

class CFileEnum : public IBoxedAppEnumVirtualFile
{
private:

	// Reference count
	LONG m_nRefCount;

	// Collection of properties
	typedef std::list<CFileProperty> TProps;
	TProps m_Props;

	TProps::const_iterator m_pos;

public:
	CFileEnum() : 
		m_nRefCount(1)
	{
	}

protected:
	virtual ~CFileEnum()
	{
		// TODO: free resources
	}

public:

	void AddProp(const CFileProperty& prop)
	{
		m_Props.push_back(prop);
	}

// IUnknown

    virtual ULONG STDMETHODCALLTYPE AddRef()
	{
		return InterlockedIncrement(&m_nRefCount);
	}

    virtual ULONG STDMETHODCALLTYPE Release()
	{
		ULONG nRefCount = InterlockedDecrement(&m_nRefCount);

		if (0 == nRefCount)
		{
			delete this;
		}

		return nRefCount;
	}

	virtual HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void** ppObject)
	{
		*ppObject = NULL;

		if (IsEqualIID(iid, IID_IUnknown))
			*ppObject = this;
		else if (IsEqualIID(iid, IID_IBoxedAppEnumVirtualFile))
			*ppObject = (IBoxedAppEnumVirtualFile*)this;

		if (NULL != *ppObject)
		{
			AddRef();
			return S_OK;
		}
		else
		{
			return E_NOINTERFACE;
		}
	}

// IBoxedAppEnumVirtualFile

    virtual HRESULT STDMETHODCALLTYPE Next( 
        /* [in] */ ULONG celt,
        /* [out][in] */ SBoxedAppVirtualFileData *pData,
        /* [out] */ ULONG *pceltFetched)
	{
		ULONG i = 0;

		for (; m_Props.end() != m_pos && i < celt; m_pos++, i++)
		{
			m_pos->CopyTo(pData[i]);
		}

		if (NULL != pceltFetched)
		{
			*pceltFetched = i;
		}

		return i == celt ? S_OK : S_FALSE;
	}
    
    virtual HRESULT STDMETHODCALLTYPE Skip( 
        /* [in] */ ULONG celt)
	{
		ULONG i = 0;

		for (; m_Props.end() != m_pos && i < celt; m_pos++, i++)
		{
		}

		return i == celt ? S_OK : S_FALSE;
	}
    
    virtual HRESULT STDMETHODCALLTYPE Reset()
	{
		m_pos = m_Props.begin();

		return S_OK;
	}
    
    virtual HRESULT STDMETHODCALLTYPE Clone( 
        /* [retval][out] */ IBoxedAppEnumVirtualFile **ppEnum)
	{
		CFileEnum* pClone = new CFileEnum(*this);

		pClone->Reset();

		*ppEnum = pClone;

		return S_OK;
	}
};


class CMyVirtualFile;

// Represents file stream
class CMyVirtualFileStream : public IBoxedAppVirtualFileStream
{
	friend class CMyVirtualFile;

private:

	// Reference count
	LONG m_nRefCount;

	// Attributes
	DWORD m_dwAttributes;

	// Creation time
	LARGE_INTEGER m_CreationTime;
	// Last access time
	LARGE_INTEGER m_LastAccessTime;
	// Last write time
	LARGE_INTEGER m_LastWriteTime;
	// Change time
	LARGE_INTEGER m_ChangeTime;

	// Memory-based stream
	IStream* m_pStream;

public:
	CMyVirtualFileStream() : 
		m_nRefCount(1), 
		m_dwAttributes(FILE_ATTRIBUTE_NORMAL), 
		m_pStream(NULL)
	{
		// Create memory-based IStream

		CreateStreamOnHGlobal(NULL, TRUE, &m_pStream);

		// Save current time

		SYSTEMTIME current_time;
		GetSystemTime(&current_time);
		FILETIME current_file_time;
		SystemTimeToFileTime(&current_time, &current_file_time);

		LARGE_INTEGER cur_time = { current_file_time.dwLowDateTime, current_file_time.dwHighDateTime };

		m_CreationTime = m_LastAccessTime = m_LastWriteTime = m_ChangeTime = cur_time;
	}

protected:
	virtual ~CMyVirtualFileStream()
	{
		// Free resources

		if (NULL != m_pStream)
		{
			m_pStream->Release();
			m_pStream = NULL;
		}
	}

public:

	LARGE_INTEGER GetSize()
	{
		STATSTG statstg;
		m_pStream->Stat(&statstg, STATFLAG_NONAME);
		LARGE_INTEGER size;
		size.QuadPart = statstg.cbSize.QuadPart;
		return size;
	}

// IUnknown

    virtual ULONG STDMETHODCALLTYPE AddRef()
	{
		return InterlockedIncrement(&m_nRefCount);
	}

    virtual ULONG STDMETHODCALLTYPE Release()
	{
		ULONG nRefCount = InterlockedDecrement(&m_nRefCount);

		if (0 == nRefCount)
		{
			delete this;
		}

		return nRefCount;
	}

	virtual HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void** ppObject)
	{
		*ppObject = NULL;

		if (IsEqualIID(iid, IID_IUnknown))
			*ppObject = this;
		else if (IsEqualIID(iid, IID_IBoxedAppVirtualFileStream))
			*ppObject = (IBoxedAppVirtualFileStream*)this;

		if (NULL != *ppObject)
		{
			AddRef();
			return S_OK;
		}
		else
		{
			return E_NOINTERFACE;
		}
	}

// IBoxedAppVirtualFileStream

    virtual HRESULT STDMETHODCALLTYPE get_Attributes( 
        /* [retval][out] */ DWORD *pdwAttributes)
	{
		*pdwAttributes = m_dwAttributes;
		return S_OK;
	}
    
    virtual HRESULT STDMETHODCALLTYPE set_Attributes( 
        /* [in] */ DWORD dwAttributes)
	{
		m_dwAttributes = dwAttributes;
		return S_OK;
	}
    
    virtual HRESULT STDMETHODCALLTYPE get_CreationTime( 
        /* [retval][out] */ LARGE_INTEGER *ptime)
	{
		*ptime = m_CreationTime;
		return S_OK;
	}
    virtual HRESULT STDMETHODCALLTYPE set_CreationTime( 
        /* [in] */ LARGE_INTEGER *ptime)
	{
		m_CreationTime = *ptime;
		return S_OK;
	}

    virtual HRESULT STDMETHODCALLTYPE get_LastWriteTime( 
        /* [retval][out] */ LARGE_INTEGER *ptime)
	{
		*ptime = m_LastWriteTime;
		return S_OK;
	}
    virtual HRESULT STDMETHODCALLTYPE set_LastWriteTime( 
        /* [in] */ LARGE_INTEGER *ptime)
	{
		m_LastWriteTime = *ptime;
		return S_OK;
	}

    virtual HRESULT STDMETHODCALLTYPE get_LastAccessTime( 
        /* [retval][out] */ LARGE_INTEGER *ptime)
	{
		*ptime = m_LastAccessTime;
		return S_OK;
	}
    virtual HRESULT STDMETHODCALLTYPE set_LastAccessTime( 
        /* [in] */ LARGE_INTEGER *ptime)
	{
		m_LastAccessTime = *ptime;
		return S_OK;
	}

    virtual HRESULT STDMETHODCALLTYPE get_ChangeTime( 
        /* [retval][out] */ LARGE_INTEGER *ptime)
	{
		*ptime = m_ChangeTime;
		return S_OK;
	}
    virtual HRESULT STDMETHODCALLTYPE set_ChangeTime( 
        /* [in] */ LARGE_INTEGER *ptime)
	{
		m_ChangeTime = *ptime;
		return S_OK;
	}
    
    virtual HRESULT STDMETHODCALLTYPE get_NewStream( 
        /* [retval][out] */ LPSTREAM *ppStream)
	{
		return m_pStream->Clone(ppStream);
	}
};

// Represents both types of file: directory and non-directory
class CMyVirtualFile : public IBoxedAppVirtualFile
{
private:

	// Reference count
	LONG m_nRefCount;

	// Attributes
	DWORD m_dwAttributes;

	// Is directory?
	BOOL m_bDir;

// For directory (for non-directory these properties are the same as default file stream has)

	// Creation time
	LARGE_INTEGER m_CreationTime;
	// Last access time
	LARGE_INTEGER m_LastAccessTime;
	// Last write time
	LARGE_INTEGER m_LastWriteTime;
	// Change time
	LARGE_INTEGER m_ChangeTime;

// For non-directory
	CMyVirtualFileStream* m_pDefaultStream;

	// File streams
	typedef std::map<std::wstring, CMyVirtualFileStream*> TFileStreams;
	TFileStreams m_FileStreams;

	// Child files
	typedef std::map<std::wstring, CMyVirtualFile*> TFiles;
	TFiles m_Files;

public:
	CMyVirtualFile(DWORD dwAttributes) : 
		m_nRefCount(1), 
		m_dwAttributes(dwAttributes)
	{
		m_bDir = 0 != (m_dwAttributes & FILE_ATTRIBUTE_DIRECTORY);

		if (!m_bDir)
		{
			m_pDefaultStream = new CMyVirtualFileStream;
		}
		else
		{
			m_pDefaultStream = NULL;

			// Save current time

			SYSTEMTIME current_time;
			GetSystemTime(&current_time);
			FILETIME current_file_time;
			SystemTimeToFileTime(&current_time, &current_file_time);

			LARGE_INTEGER cur_time = { current_file_time.dwLowDateTime, current_file_time.dwHighDateTime };

			m_CreationTime = m_LastAccessTime = m_LastWriteTime = m_ChangeTime = cur_time;
		}
	}

protected:

	virtual ~CMyVirtualFile()
	{
		for (TFileStreams::iterator it = m_FileStreams.begin(); m_FileStreams.end() != it; it++)
		{
			it->second->Release();
		}

		m_FileStreams.clear();

		for (TFiles::iterator it = m_Files.begin(); m_Files.end() != it; it++)
		{
			it->second->Release();
		}

		m_Files.clear();

		if (NULL != m_pDefaultStream)
		{
			m_pDefaultStream->Release();
			m_pDefaultStream = NULL;
		}
	}

public:

// IUnknown

    virtual ULONG STDMETHODCALLTYPE AddRef()
	{
		return InterlockedIncrement(&m_nRefCount);
	}

    virtual ULONG STDMETHODCALLTYPE Release()
	{
		ULONG nRefCount = InterlockedDecrement(&m_nRefCount);

		if (0 == nRefCount)
		{
			delete this;
		}

		return nRefCount;
	}

	virtual HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void** ppObject)
	{
		*ppObject = NULL;

		if (IsEqualIID(iid, IID_IUnknown))
			*ppObject = this;
		else if (IsEqualIID(iid, IID_IBoxedAppVirtualFile))
			*ppObject = (IBoxedAppVirtualFile*)this;

		if (NULL != *ppObject)
		{
			AddRef();
			return S_OK;
		}
		else
		{
			return E_NOINTERFACE;
		}
	}

// IBoxedAppVirtualFile

    virtual HRESULT STDMETHODCALLTYPE get_Attributes( 
        /* [retval][out] */ DWORD *pdwAttributes)
	{
		if (!m_bDir)
			return m_pDefaultStream->get_Attributes(pdwAttributes);
		else
		{
			*pdwAttributes = m_dwAttributes;
			return S_OK;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE set_Attributes( 
        /* [in] */ DWORD dwAttributes)
	{
		if (!m_bDir)
		{
			return m_pDefaultStream->set_Attributes(dwAttributes);
		}
		else
		{
			m_dwAttributes = dwAttributes;
			return S_OK;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE get_CreationTime( 
        /* [retval][out] */ LARGE_INTEGER *ptime)
	{
		if (!m_bDir)
		{
			return m_pDefaultStream->get_CreationTime(ptime);
		}
		else
		{
			*ptime = m_CreationTime;
			return S_OK;
		}
	}
    virtual HRESULT STDMETHODCALLTYPE set_CreationTime( 
        /* [in] */ LARGE_INTEGER *ptime)
	{
		if (!m_bDir)
		{
			return m_pDefaultStream->set_CreationTime(ptime);
		}
		else
		{
			m_CreationTime = *ptime;
			return S_OK;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE get_LastAccessTime( 
        /* [retval][out] */ LARGE_INTEGER *ptime)
	{
		if (!m_bDir)
		{
			return m_pDefaultStream->get_LastAccessTime(ptime);
		}
		else
		{
			*ptime = m_LastAccessTime;
			return S_OK;
		}
	}
    virtual HRESULT STDMETHODCALLTYPE set_LastAccessTime( 
        /* [in] */ LARGE_INTEGER *ptime)
	{
		if (!m_bDir)
		{
			return m_pDefaultStream->set_LastAccessTime(ptime);
		}
		else
		{
			m_LastAccessTime = *ptime;
			return S_OK;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE get_LastWriteTime( 
        /* [retval][out] */ LARGE_INTEGER *ptime)
	{
		if (!m_bDir)
		{
			return m_pDefaultStream->get_LastWriteTime(ptime);
		}
		else
		{
			*ptime = m_LastWriteTime;
			return S_OK;
		}
	}
    virtual HRESULT STDMETHODCALLTYPE set_LastWriteTime( 
        /* [in] */ LARGE_INTEGER *ptime)
	{
		if (!m_bDir)
		{
			return m_pDefaultStream->set_LastWriteTime(ptime);
		}
		else
		{
			m_LastWriteTime = *ptime;
			return S_OK;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE get_ChangeTime( 
        /* [retval][out] */ LARGE_INTEGER *ptime)
	{
		if (!m_bDir)
		{
			return m_pDefaultStream->get_ChangeTime(ptime);
		}
		else
		{
			*ptime = m_ChangeTime;
			return S_OK;
		}
	}
    virtual HRESULT STDMETHODCALLTYPE set_ChangeTime( 
        /* [in] */ LARGE_INTEGER *ptime)
	{
		if (!m_bDir)
		{
			return m_pDefaultStream->set_ChangeTime(ptime);
		}
		else
		{
			m_ChangeTime = *ptime;
			return S_OK;
		}
	}

	virtual HRESULT STDMETHODCALLTYPE createFileStream( 
        /* [in] */ LPCWSTR szName,
        /* [in] */ DWORD dwDesiredAccess,
        /* [in] */ DWORD dwAttributes,
        /* [in] */ DWORD dwShareAccess,
        /* [in] */ DWORD dwCreateOptions,
        /* [retval][out] */ IBoxedAppVirtualFileStream **ppVirtualFileStream)
	{
		if (m_FileStreams.end() == m_FileStreams.find(szName))
			// No such stream => create it
		{
			CMyVirtualFileStream* pMyVirtualFileStream = new CMyVirtualFileStream();

			m_FileStreams[szName] = pMyVirtualFileStream;

			pMyVirtualFileStream->QueryInterface(IID_IBoxedAppVirtualFileStream, (void**)ppVirtualFileStream);

			return S_OK;
		}
		else
		{
			// TODO: use another return code
			return E_FAIL;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE openFileStream( 
        /* [in] */ LPCWSTR szName,
        /* [in] */ DWORD dwDesiredAccess,
        /* [in] */ DWORD dwShareAccess,
        /* [in] */ DWORD dwOpenOptions,
        /* [retval][out] */ IBoxedAppVirtualFileStream **ppVirtualFileStream)
	{
		if (!m_bDir && (NULL == szName || 0 == szName[0]))
			// Default stream for non-directory file
		{
			return m_pDefaultStream->QueryInterface(IID_IBoxedAppVirtualFileStream, (void**)ppVirtualFileStream);
		}

		TFileStreams::const_iterator it = m_FileStreams.find(szName);

		if (m_FileStreams.end() == it)
			// No such stream => return error
		{
			// TODO: use another return code
			return E_FAIL;
		}
		else
		{
			IBoxedAppVirtualFileStream* pMyVirtualFileStream = it->second;

			pMyVirtualFileStream->QueryInterface(IID_IBoxedAppVirtualFileStream, (void**)ppVirtualFileStream);

			return S_OK;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE renameFileStream( 
        /* [in] */ LPCWSTR szOldName,
        /* [in] */ LPCWSTR szNewName)
	{
		TFileStreams::iterator it = m_FileStreams.find(szOldName);

		if (m_FileStreams.end() == it)
			// No such stream => return error
		{
			// TODO: use another return code
			return E_FAIL;
		}
		else
		{
			m_FileStreams[szNewName] = it->second;
			m_FileStreams.erase(it);

			return S_OK;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE deleteFileStream( 
        /* [in] */ LPCWSTR szName)
	{
		TFileStreams::iterator it = m_FileStreams.find(szName);

		if (m_FileStreams.end() == it)
			// No such stream => return error
		{
			// TODO: use another return code
			return E_FAIL;
		}
		else
		{
			it->second->Release();

			m_FileStreams.erase(it);

			return S_OK;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE enumFileStreams( 
        /* [retval][out] */ IBoxedAppEnumVirtualFile **ppEnumVirtualStream)
	{
		CFileEnum* pFileEnum = new CFileEnum();

		for (TFileStreams::iterator it = m_FileStreams.begin(); m_FileStreams.end() != it; it++)
		{
			CMyVirtualFileStream* pFileStream = it->second;

			DWORD dwAttributes;
			pFileStream->get_Attributes(&dwAttributes);

			pFileEnum->AddProp(
				CFileProperty(
						it->first.c_str(), 
						pFileStream->GetSize(), 
						dwAttributes, 
						pFileStream->m_CreationTime, 
						pFileStream->m_LastWriteTime, 
						pFileStream->m_LastAccessTime, 
						pFileStream->m_ChangeTime
					)
				);
		}

		pFileEnum->Reset();

		*ppEnumVirtualStream = pFileEnum;

		return S_OK;
	}

    virtual HRESULT STDMETHODCALLTYPE createFile( 
        /* [in] */ LPCWSTR szName,
        /* [in] */ DWORD dwDesiredAccess,
        /* [in] */ DWORD dwAttributes,
        /* [in] */ DWORD dwShareAccess,
        /* [in] */ DWORD dwCreateOptions,
        /* [retval][out] */ IBoxedAppVirtualFile **ppVirtualFile)
	{
		if (m_Files.end() == m_Files.find(szName))
			// No such stream => create it
		{
			CMyVirtualFile* pMyVirtualFile = new CMyVirtualFile(dwAttributes);

			m_Files[szName] = pMyVirtualFile;

			pMyVirtualFile->QueryInterface(IID_IBoxedAppVirtualFile, (void**)ppVirtualFile);

			return S_OK;
		}
		else
		{
			// TODO: use another return code
			return E_FAIL;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE openFile( 
        /* [in] */ LPCWSTR szName,
        /* [in] */ DWORD dwDesiredAccess,
        /* [in] */ DWORD dwShareAccess,
        /* [in] */ DWORD dwOpenOptions,
        /* [retval][out] */ IBoxedAppVirtualFile **ppVirtualFile)
	{
		TFiles::const_iterator it = m_Files.find(szName);

		if (m_Files.end() == it)
			// No such stream => return error
		{
			// TODO: use another return code
			return E_FAIL;
		}
		else
		{
			IBoxedAppVirtualFile* pMyVirtualFile = it->second;

			pMyVirtualFile->QueryInterface(IID_IBoxedAppVirtualFile, (void**)ppVirtualFile);

			return S_OK;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE renameFile( 
        /* [in] */ LPCWSTR szOldName,
        /* [in] */ LPCWSTR szNewName)
	{
		TFiles::iterator it = m_Files.find(szOldName);

		if (m_Files.end() == it)
			// No such stream => return error
		{
			// TODO: use another return code
			return E_FAIL;
		}
		else
		{
			m_Files[szNewName] = it->second;
			m_Files.erase(it);

			return S_OK;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE deleteFile( 
        /* [in] */ LPCWSTR szName)
	{
		TFiles::iterator it = m_Files.find(szName);

		if (m_Files.end() == it)
			// No such stream => return error
		{
			// TODO: use another return code
			return E_FAIL;
		}
		else
		{
			it->second->Release();

			m_Files.erase(it);

			return S_OK;
		}
	}

    virtual HRESULT STDMETHODCALLTYPE enumFiles( 
        /* [retval][out] */ IBoxedAppEnumVirtualFile **ppEnumVirtualStream)
	{
		CFileEnum* pFileEnum = new CFileEnum();

		LARGE_INTEGER liZero = { 0 };

		for (TFiles::iterator it = m_Files.begin(); m_Files.end() != it; it++)
		{
			CMyVirtualFile* pFile = it->second;

			DWORD dwAttributes;
			pFile->get_Attributes(&dwAttributes);

			LARGE_INTEGER CreationTime, LastWriteTime, LastAccessTime, ChangeTime;
			pFile->get_CreationTime(&CreationTime);
			pFile->get_LastWriteTime(&LastWriteTime);
			pFile->get_LastAccessTime(&LastAccessTime);
			pFile->get_ChangeTime(&ChangeTime);

			pFileEnum->AddProp(
				CFileProperty(
						it->first.c_str(), 
						pFile->m_bDir ? liZero : pFile->m_pDefaultStream->GetSize(), 
						dwAttributes, 
						CreationTime, 
						LastWriteTime, 
						LastAccessTime, 
						ChangeTime
					)
				);
		}

		pFileEnum->Reset();

		*ppEnumVirtualStream = pFileEnum;

		return S_OK;
	}
};

HRESULT CreateMyVirtualFile(DWORD dwAttributes, IBoxedAppVirtualFile** ppFile)
{
	CMyVirtualFile* pMyVirtualFile = new CMyVirtualFile(dwAttributes);

	if (NULL == pMyVirtualFile)
	{
		return E_OUTOFMEMORY;
	}
	else
	{
		*ppFile = pMyVirtualFile;

		return S_OK;
	}
}
